# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from abc import abstractmethod
from typing import Any, Optional

from azure.ai.ml import Input

from .automl_job import AutoMLJob


class AutoMLVertical(AutoMLJob):
    """Abstract class for AutoML verticals.

    :param task_type: The type of task to run. Possible values include: "classification", "regression", "forecasting".
    :type task_type: str
    :param training_data: Training data input
    :type training_data: Input
    :param validation_data: Validation data input
    :type validation_data: Input
    :param test_data: Test data input, defaults to None
    :type test_data: typing.Optional[Input]
    :raises ValueError: If task_type is not one of "classification", "regression", "forecasting".
    :raises ValueError: If training_data is not of type Input.
    :raises ValueError: If validation_data is not of type Input.
    :raises ValueError: If test_data is not of type Input.
    """

    @abstractmethod
    def __init__(
        self,
        task_type: str,
        training_data: Input,
        validation_data: Input,
        test_data: Optional[Input] = None,
        **kwargs: Any
    ) -> None:
        """Initialize AutoMLVertical.

        Constructor for AutoMLVertical.

        :param task_type: The type of task to run. Possible values include: "classification", "regression"
            , "forecasting".
        :type task_type: str
        :param training_data: Training data input
        :type training_data: Input
        :param validation_data: Validation data input
        :type validation_data: Input
        :param test_data: Test data input, defaults to None
        :type test_data: typing.Optional[Input]
        :raises ValueError: If task_type is not one of "classification", "regression", "forecasting".
        :raises ValueError: If training_data is not of type Input.
        :raises ValueError: If validation_data is not of type Input.
        :raises ValueError: If test_data is not of type Input.
        """
        self._task_type = task_type
        self.training_data = training_data
        self.validation_data = validation_data
        self.test_data = test_data  # type: ignore
        super().__init__(**kwargs)

    @property
    def task_type(self) -> str:
        """Get task type.

        :return: The type of task to run. Possible values include: "classification", "regression", "forecasting".
        :rtype: str
        """
        return self._task_type

    @task_type.setter
    def task_type(self, task_type: str) -> None:
        """Set task type.

        :param task_type: The type of task to run. Possible values include: "classification", "regression"
            , "forecasting".
        :type task_type: str
        """
        self._task_type = task_type

    @property
    def training_data(self) -> Input:
        """Get training data.

        :return: Training data input
        :rtype: Input
        """
        return self._training_data

    @training_data.setter
    def training_data(self, training_data: Input) -> None:
        """Set training data.

        :param training_data: Training data input
        :type training_data: Input
        """
        self._training_data = training_data

    @property
    def validation_data(self) -> Input:
        """Get validation data.

        :return: Validation data input
        :rtype: Input
        """
        return self._validation_data

    @validation_data.setter
    def validation_data(self, validation_data: Input) -> None:
        """Set validation data.

        :param validation_data: Validation data input
        :type validation_data: Input
        """
        self._validation_data = validation_data

    @property
    def test_data(self) -> Input:
        """Get test data.

        :return: Test data input
        :rtype: Input
        """
        return self._test_data

    @test_data.setter
    def test_data(self, test_data: Input) -> None:
        """Set test data.

        :param test_data: Test data input
        :type test_data: Input
        """
        self._test_data = test_data



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

from typing import Any, Dict, List, Optional, Union

from azure.ai.ml._restclient.v2023_04_01_preview.models import AutoMLJob as RestAutoMLJob
from azure.ai.ml._restclient.v2023_04_01_preview.models import Forecasting as RestForecasting
from azure.ai.ml._restclient.v2023_04_01_preview.models import ForecastingPrimaryMetrics, JobBase, TaskType
from azure.ai.ml._utils.utils import camel_to_snake, is_data_binding_expression
from azure.ai.ml.constants import TabularTrainingMode
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY
from azure.ai.ml.constants._job.automl import AutoMLConstants
from azure.ai.ml.entities._credentials import _BaseJobIdentityConfiguration
from azure.ai.ml.entities._job._input_output_helpers import from_rest_data_outputs, to_rest_data_outputs
from azure.ai.ml.entities._job.automl.stack_ensemble_settings import StackEnsembleSettings
from azure.ai.ml.entities._job.automl.tabular.automl_tabular import AutoMLTabular
from azure.ai.ml.entities._job.automl.tabular.featurization_settings import TabularFeaturizationSettings
from azure.ai.ml.entities._job.automl.tabular.forecasting_settings import ForecastingSettings
from azure.ai.ml.entities._job.automl.tabular.limit_settings import TabularLimitSettings
from azure.ai.ml.entities._job.automl.training_settings import ForecastingTrainingSettings
from azure.ai.ml.entities._util import load_from_dict


class ForecastingJob(AutoMLTabular):
    """
    Configuration for AutoML Forecasting Task.

    :param primary_metric: The primary metric to use for model selection.
    :type primary_metric: Optional[str]
    :param forecasting_settings: The settings for the forecasting task.
    :type forecasting_settings:
        Optional[~azure.ai.ml.automl.ForecastingSettings]
    :param kwargs: Job-specific arguments
    :type kwargs: Dict[str, Any]
    """

    _DEFAULT_PRIMARY_METRIC = ForecastingPrimaryMetrics.NORMALIZED_ROOT_MEAN_SQUARED_ERROR

    def __init__(
        self,
        *,
        primary_metric: Optional[str] = None,
        forecasting_settings: Optional[ForecastingSettings] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a new AutoML Forecasting task."""
        # Extract any task specific settings
        featurization = kwargs.pop("featurization", None)
        limits = kwargs.pop("limits", None)
        training = kwargs.pop("training", None)

        super().__init__(
            task_type=TaskType.FORECASTING,
            featurization=featurization,
            limits=limits,
            training=training,
            **kwargs,
        )

        self.primary_metric = primary_metric or ForecastingJob._DEFAULT_PRIMARY_METRIC
        self._forecasting_settings = forecasting_settings

    @property
    def primary_metric(self) -> Optional[str]:
        """
        Return the primary metric to use for model selection.

        :return: The primary metric for model selection.
        :rtype: Optional[str]
        """
        return self._primary_metric

    @primary_metric.setter
    def primary_metric(self, value: Union[str, ForecastingPrimaryMetrics]) -> None:
        """
        Set the primary metric to use for model selection.

        :param value: The primary metric for model selection.
        :type: Union[str, ~azure.ai.ml.automl.ForecastingPrimaryMetrics]
        """
        if is_data_binding_expression(str(value), ["parent"]):
            self._primary_metric = value
            return
        self._primary_metric = (
            ForecastingJob._DEFAULT_PRIMARY_METRIC
            if value is None
            else ForecastingPrimaryMetrics[camel_to_snake(value).upper()]
        )

    @property  # type: ignore
    def training(self) -> ForecastingTrainingSettings:
        """
        Return the forecast training settings.

        :return: training settings.
        :rtype: ~azure.ai.ml.automl.ForecastingTrainingSettings
        """
        return self._training or ForecastingTrainingSettings()

    @training.setter
    def training(self, value: Union[Dict, ForecastingTrainingSettings]) -> None:  # pylint: disable=unused-argument
        ...

    @property
    def forecasting_settings(self) -> Optional[ForecastingSettings]:
        """
        Return the forecast settings.

        :return: forecast settings.
        :rtype: ~azure.ai.ml.automl.ForecastingSettings
        """
        return self._forecasting_settings

    def set_forecast_settings(
        self,
        *,
        time_column_name: Optional[str] = None,
        forecast_horizon: Optional[Union[str, int]] = None,
        time_series_id_column_names: Optional[Union[str, List[str]]] = None,
        target_lags: Optional[Union[str, int, List[int]]] = None,
        feature_lags: Optional[str] = None,
        target_rolling_window_size: Optional[Union[str, int]] = None,
        country_or_region_for_holidays: Optional[str] = None,
        use_stl: Optional[str] = None,
        seasonality: Optional[Union[str, int]] = None,
        short_series_handling_config: Optional[str] = None,
        frequency: Optional[str] = None,
        target_aggregate_function: Optional[str] = None,
        cv_step_size: Optional[int] = None,
        features_unknown_at_forecast_time: Optional[Union[str, List[str]]] = None,
    ) -> None:
        """Manage parameters used by forecasting tasks.

        :keyword time_column_name:
            The name of the time column. This parameter is required when forecasting to specify the datetime
            column in the input data used for building the time series and inferring its frequency.
        :paramtype time_column_name: Optional[str]
        :keyword forecast_horizon:
            The desired maximum forecast horizon in units of time-series frequency. The default value is 1.

            Units are based on the time interval of your training data, e.g., monthly, weekly that the forecaster
            should predict out. When task type is forecasting, this parameter is required. For more information on
            setting forecasting parameters, see `Auto-train a time-series forecast model <https://learn.microsoft.com/
            azure/machine-learning/how-to-auto-train-forecast>`_.
        :type forecast_horizon: Optional[Union[int, str]]
        :keyword time_series_id_column_names:
            The names of columns used to group a time series.
            It can be used to create multiple series. If time series id column names is not defined or
            the identifier columns specified do not identify all the series in the dataset, the time series identifiers
            will be automatically created for your data set.
        :paramtype time_series_id_column_names: Optional[Union[str, List[str]]]
        :keyword target_lags: The number of past periods to lag from the target column. By default the lags are turned
            off.

            When forecasting, this parameter represents the number of rows to lag the target values based
            on the frequency of the data. This is represented as a list or single integer. Lag should be used
            when the relationship between the independent variables and dependent variable do not match up or
            correlate by default. For example, when trying to forecast demand for a product, the demand in any
            month may depend on the price of specific commodities 3 months prior. In this example, you may want
            to lag the target (demand) negatively by 3 months so that the model is training on the correct
            relationship. For more information, see `Auto-train a time-series forecast model
            <https://learn.microsoft.com/azure/machine-learning/how-to-auto-train-forecast>`_.

            **Note on auto detection of target lags and rolling window size.
            Please see the corresponding comments in the rolling window section.**
            We use the next algorithm to detect the optimal target lag and rolling window size.

            #. Estimate the maximum lag order for the look back feature selection. In our case it is the number of
               periods till the next date frequency granularity i.e. if frequency is daily, it will be a week (7),
               if it is a week, it will be month (4). That values multiplied by two is the largest
               possible values of lags/rolling windows. In our examples, we will consider the maximum lag
               order of 14 and 8 respectively).
            #. Create a de-seasonalized series by adding trend and residual components. This will be used
               in the next step.
            #. Estimate the PACF - Partial Auto Correlation Function on the on the data from (2)
               and search for points, where the auto correlation is significant i.e. its absolute
               value is more then 1.96/square_root(maximal lag value), which correspond to significance of 95%.
            #. If all points are significant, we consider it being strong seasonality
               and do not create look back features.
            #. We scan the PACF values from the beginning and the value before the first insignificant
               auto correlation will designate the lag. If first significant element (value correlate with
               itself) is followed by insignificant, the lag will be 0 and we will not use look back features.

        :type target_lags: Optional[Union[str, int, List[int]]]
        :keyword feature_lags: Flag for generating lags for the numeric features with 'auto' or None.
        :paramtype feature_lags: Optional[str]
        :keyword target_rolling_window_size: The number of past periods used to create a rolling window average of the
            target column.

            When forecasting, this parameter represents `n` historical periods to use to generate forecasted values,
            <= training set size. If omitted, `n` is the full training set size. Specify this parameter
            when you only want to consider a certain amount of history when training the model.
            If set to 'auto', rolling window will be estimated as the last
            value where the PACF is more then the significance threshold. Please see target_lags section for details.
        :paramtype target_rolling_window_size: Optional[Union[str, int]]
        :keyword country_or_region_for_holidays: The country/region used to generate holiday features.
            These should be ISO 3166 two-letter country/region codes, for example 'US' or 'GB'.
        :paramtype country_or_region_for_holidays: Optional[str]
        :keyword use_stl: Configure STL Decomposition of the time-series target column.
            use_stl can take three values: None (default) - no stl decomposition, 'season' - only generate
            season component and season_trend - generate both season and trend components.
        :type use_stl: Optional[str]
        :keyword seasonality: Set time series seasonality as an integer multiple of the series frequency.
            If seasonality is set to 'auto', it will be inferred.
            If set to None, the time series is assumed non-seasonal which is equivalent to seasonality=1.
        :paramtype seasonality: Optional[Union[int, str]
        :keyword short_series_handling_config:
            The parameter defining how if AutoML should handle short time series.

            Possible values: 'auto' (default), 'pad', 'drop' and None.

            * **auto** short series will be padded if there are no long series,
                otherwise short series will be dropped.
            * **pad** all the short series will be padded.
            * **drop**  all the short series will be dropped".
            * **None** the short series will not be modified.

            If set to 'pad', the table will be padded with the zeroes and
            empty values for the regressors and random values for target with the mean
            equal to target value median for given time series id. If median is more or equal
            to zero, the minimal padded value will be clipped by zero:
            Input:

            +------------+---------------+----------+--------+
            | Date       | numeric_value | string   | target |
            +============+===============+==========+========+
            | 2020-01-01 | 23            | green    | 55     |
            +------------+---------------+----------+--------+

            Output assuming minimal number of values is four:

            +------------+---------------+----------+--------+
            | Date       | numeric_value | string   | target |
            +============+===============+==========+========+
            | 2019-12-29 | 0             | NA       | 55.1   |
            +------------+---------------+----------+--------+
            | 2019-12-30 | 0             | NA       | 55.6   |
            +------------+---------------+----------+--------+
            | 2019-12-31 | 0             | NA       | 54.5   |
            +------------+---------------+----------+--------+
            | 2020-01-01 | 23            | green    | 55     |
            +------------+---------------+----------+--------+

            **Note:** We have two parameters short_series_handling_configuration and
            legacy short_series_handling. When both parameters are set we are
            synchronize them as shown in the table below (short_series_handling_configuration and
            short_series_handling for brevity are marked as handling_configuration and handling
            respectively).

            +------------+--------------------------+----------------------+-----------------------------+
            | | handling | | handling               | | resulting          | | resulting                 |
            |            | | configuration          | | handling           | | handling                  |
            |            |                          |                      | | configuration             |
            +============+==========================+======================+=============================+
            | True       | auto                     | True                 | auto                        |
            +------------+--------------------------+----------------------+-----------------------------+
            | True       | pad                      | True                 | auto                        |
            +------------+--------------------------+----------------------+-----------------------------+
            | True       | drop                     | True                 | auto                        |
            +------------+--------------------------+----------------------+-----------------------------+
            | True       | None                     | False                | None                        |
            +------------+--------------------------+----------------------+-----------------------------+
            | False      | auto                     | False                | None                        |
            +------------+--------------------------+----------------------+-----------------------------+
            | False      | pad                      | False                | None                        |
            +------------+--------------------------+----------------------+-----------------------------+
            | False      | drop                     | False                | None                        |
            +------------+--------------------------+----------------------+-----------------------------+
            | False      | None                     | False                | None                        |
            +------------+--------------------------+----------------------+-----------------------------+

        :type short_series_handling_config: Optional[str]
        :keyword frequency: Forecast frequency.

            When forecasting, this parameter represents the period with which the forecast is desired,
            for example daily, weekly, yearly, etc. The forecast frequency is dataset frequency by default.
            You can optionally set it to greater (but not lesser) than dataset frequency.
            We'll aggregate the data and generate the results at forecast frequency. For example,
            for daily data, you can set the frequency to be daily, weekly or monthly, but not hourly.
            The frequency needs to be a pandas offset alias.
            Please refer to pandas documentation for more information:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
        :type frequency: Optional[str]
        :keyword target_aggregate_function: The function to be used to aggregate the time series target
            column to conform to a user specified frequency. If the target_aggregation_function is set,
            but the freq parameter is not set, the error is raised. The possible target aggregation
            functions are: "sum", "max", "min" and "mean".

                * The target column values are aggregated based on the specified operation.
                  Typically, sum is appropriate for most scenarios.
                * Numerical predictor columns in your data are aggregated by sum, mean, minimum value,
                  and maximum value. As a result, automated ML generates new columns suffixed with the
                  aggregation function name and applies the selected aggregate operation.
                * For categorical predictor columns, the data is aggregated by mode,
                  the most prominent category in the window.
                * Date predictor columns are aggregated by minimum value, maximum value and mode.

                +----------------+-------------------------------+--------------------------------------+
                |     | freq     | | target_aggregation_function | | Data regularity                    |
                |                |                               | | fixing mechanism                   |
                +================+===============================+======================================+
                | None (Default) | None (Default)                | | The aggregation                    |
                |                |                               | | is not applied.                    |
                |                |                               | | If the valid                       |
                |                |                               | | frequency can                      |
                |                |                               | | not be                             |
                |                |                               | | determined                         |
                |                |                               | | the error                          |
                |                |                               | | will be raised.                    |
                +----------------+-------------------------------+--------------------------------------+
                | Some Value     | None (Default)                | | The aggregation                    |
                |                |                               | | is not applied.                    |
                |                |                               | | If the number                      |
                |                |                               | | of data points                     |
                |                |                               | | compliant to                       |
                |                |                               | | given frequency                    |
                |                |                               | | grid is                            |
                |                |                               | | less then 90%                      |
                |                |                               | | these points                       |
                |                |                               | | will be                            |
                |                |                               | | removed,                           |
                |                |                               | | otherwise                          |
                |                |                               | | the error will                     |
                |                |                               | | be raised.                         |
                +----------------+-------------------------------+--------------------------------------+
                | None (Default) | Aggregation function          | | The error about                    |
                |                |                               | | missing                            |
                |                |                               | | frequency                          |
                |                |                               | | parameter is                       |
                |                |                               | | raised.                            |
                +----------------+-------------------------------+--------------------------------------+
                | Some Value     | Aggregation function          | | Aggregate to                       |
                |                |                               | | frequency using                    |
                |                |                               | | provided                           |
                |                |                               | | aggregation                        |
                |                |                               | | function.                          |
                +----------------+-------------------------------+--------------------------------------+

        :type target_aggregate_function: Optional[str]
        :keyword cv_step_size: Number of periods between the origin_time of one CV fold and the next fold.
            For example, if `n_step` = 3 for daily data, the origin time for each fold will be three days apart.
        :paramtype cv_step_size: Optional[int]
        :keyword features_unknown_at_forecast_time: The feature columns that are available for training but
            unknown at the time of forecast/inference. If features_unknown_at_forecast_time is set to an empty
            list, it is assumed that all the feature columns in the dataset are known at inference time. If this
            parameter is not set the support for future features is not enabled.
        :paramtype features_unknown_at_forecast_time: Optional[Union[str, List[str]]]
        """
        self._forecasting_settings = self._forecasting_settings or ForecastingSettings()

        self._forecasting_settings.country_or_region_for_holidays = (
            country_or_region_for_holidays
            if country_or_region_for_holidays is not None
            else self._forecasting_settings.country_or_region_for_holidays
        )
        self._forecasting_settings.cv_step_size = (
            cv_step_size if cv_step_size is not None else self._forecasting_settings.cv_step_size
        )
        self._forecasting_settings.forecast_horizon = (
            forecast_horizon if forecast_horizon is not None else self._forecasting_settings.forecast_horizon
        )
        self._forecasting_settings.target_lags = (
            target_lags if target_lags is not None else self._forecasting_settings.target_lags
        )
        self._forecasting_settings.target_rolling_window_size = (
            target_rolling_window_size
            if target_rolling_window_size is not None
            else self._forecasting_settings.target_rolling_window_size
        )
        self._forecasting_settings.frequency = (
            frequency if frequency is not None else self._forecasting_settings.frequency
        )
        self._forecasting_settings.feature_lags = (
            feature_lags if feature_lags is not None else self._forecasting_settings.feature_lags
        )
        self._forecasting_settings.seasonality = (
            seasonality if seasonality is not None else self._forecasting_settings.seasonality
        )
        self._forecasting_settings.use_stl = use_stl if use_stl is not None else self._forecasting_settings.use_stl
        self._forecasting_settings.short_series_handling_config = (
            short_series_handling_config
            if short_series_handling_config is not None
            else self._forecasting_settings.short_series_handling_config
        )
        self._forecasting_settings.target_aggregate_function = (
            target_aggregate_function
            if target_aggregate_function is not None
            else self._forecasting_settings.target_aggregate_function
        )
        self._forecasting_settings.time_column_name = (
            time_column_name if time_column_name is not None else self._forecasting_settings.time_column_name
        )
        self._forecasting_settings.time_series_id_column_names = (
            time_series_id_column_names
            if time_series_id_column_names is not None
            else self._forecasting_settings.time_series_id_column_names
        )
        self._forecasting_settings.features_unknown_at_forecast_time = (
            features_unknown_at_forecast_time
            if features_unknown_at_forecast_time is not None
            else self._forecasting_settings.features_unknown_at_forecast_time
        )

    # override
    def set_training(
        self,
        *,
        enable_onnx_compatible_models: Optional[bool] = None,
        enable_dnn_training: Optional[bool] = None,
        enable_model_explainability: Optional[bool] = None,
        enable_stack_ensemble: Optional[bool] = None,
        enable_vote_ensemble: Optional[bool] = None,
        stack_ensemble_settings: Optional[StackEnsembleSettings] = None,
        ensemble_model_download_timeout: Optional[int] = None,
        allowed_training_algorithms: Optional[List[str]] = None,
        blocked_training_algorithms: Optional[List[str]] = None,
        training_mode: Optional[Union[str, TabularTrainingMode]] = None,
    ) -> None:
        """
        The method to configure forecast training related settings.

        :keyword enable_onnx_compatible_models:
            Whether to enable or disable enforcing the ONNX-compatible models.
            The default is False. For more information about Open Neural Network Exchange (ONNX) and Azure Machine
            Learning, see this `article <https://learn.microsoft.com/azure/machine-learning/concept-onnx>`__.
        :type enable_onnx_compatible: Optional[bool]
        :keyword enable_dnn_training:
            Whether to include DNN based models during model selection.
            However, the default is True for DNN NLP tasks, and it's False for all other AutoML tasks.
        :paramtype enable_dnn_training: Optional[bool]
        :keyword enable_model_explainability:
            Whether to enable explaining the best AutoML model at the end of all AutoML training iterations.
            For more information, see `Interpretability: model explanations in automated machine learning
            <https://learn.microsoft.com/azure/machine-learning/how-to-machine-learning-interpretability-automl>`__.
            , defaults to None
        :type enable_model_explainability: Optional[bool]
        :keyword enable_stack_ensemble:
            Whether to enable/disable StackEnsemble iteration.
            If `enable_onnx_compatible_models` flag is being set, then StackEnsemble iteration will be disabled.
            Similarly, for Timeseries tasks, StackEnsemble iteration will be disabled by default, to avoid risks of
            overfitting due to small training set used in fitting the meta learner.
            For more information about ensembles, see `Ensemble configuration
            <https://learn.microsoft.com/azure/machine-learning/how-to-configure-auto-train#ensemble>`__
            , defaults to None
        :type enable_stack_ensemble: Optional[bool]
        :keyword enable_vote_ensemble:
            Whether to enable/disable VotingEnsemble iteration.
            For more information about ensembles, see `Ensemble configuration
            <https://learn.microsoft.com/azure/machine-learning/how-to-configure-auto-train#ensemble>`__
            , defaults to None
        :type enable_vote_ensemble: Optional[bool]
        :keyword stack_ensemble_settings:
            Settings for StackEnsemble iteration, defaults to None
        :paramtype stack_ensemble_settings: Optional[StackEnsembleSettings]
        :keyword ensemble_model_download_timeout:
            During VotingEnsemble and StackEnsemble model generation,
            multiple fitted models from the previous child runs are downloaded. Configure this parameter with a
            higher value than 300 secs, if more time is needed, defaults to None
        :paramtype ensemble_model_download_timeout: Optional[int]
        :keyword allowed_training_algorithms:
            A list of model names to search for an experiment. If not specified,
            then all models supported for the task are used minus any specified in ``blocked_training_algorithms``
            or deprecated TensorFlow models, defaults to None
        :paramtype allowed_training_algorithms: Optional[List[str]]
        :keyword blocked_training_algorithms:
            A list of algorithms to ignore for an experiment, defaults to None
        :paramtype blocked_training_algorithms: Optional[List[str]]
        :keyword training_mode:
            [Experimental] The training mode to use.
            The possible values are-

            * distributed- enables distributed training for supported algorithms.

            * non_distributed- disables distributed training.

            * auto- Currently, it is same as non_distributed. In future, this might change.

            Note: This parameter is in public preview and may change in future.
        :type training_mode: Optional[Union[~azure.ai.ml.constants.TabularTrainingMode, str]]
        """
        super().set_training(
            enable_onnx_compatible_models=enable_onnx_compatible_models,
            enable_dnn_training=enable_dnn_training,
            enable_model_explainability=enable_model_explainability,
            enable_stack_ensemble=enable_stack_ensemble,
            enable_vote_ensemble=enable_vote_ensemble,
            stack_ensemble_settings=stack_ensemble_settings,
            ensemble_model_download_timeout=ensemble_model_download_timeout,
            allowed_training_algorithms=allowed_training_algorithms,
            blocked_training_algorithms=blocked_training_algorithms,
            training_mode=training_mode,
        )

        # Disable stack ensemble by default, since it is currently not supported for forecasting tasks
        if enable_stack_ensemble is None:
            if self._training is not None:
                self._training.enable_stack_ensemble = False

    def _to_rest_object(self) -> JobBase:
        if self._forecasting_settings is not None:
            forecasting_task = RestForecasting(
                target_column_name=self.target_column_name,
                training_data=self.training_data,
                validation_data=self.validation_data,
                validation_data_size=self.validation_data_size,
                weight_column_name=self.weight_column_name,
                cv_split_column_names=self.cv_split_column_names,
                n_cross_validations=self.n_cross_validations,
                test_data=self.test_data,
                test_data_size=self.test_data_size,
                featurization_settings=self._featurization._to_rest_object() if self._featurization else None,
                limit_settings=self._limits._to_rest_object() if self._limits else None,
                training_settings=self._training._to_rest_object() if self._training else None,
                primary_metric=self.primary_metric,
                log_verbosity=self.log_verbosity,
                forecasting_settings=self._forecasting_settings._to_rest_object(),
            )
        else:
            forecasting_task = RestForecasting(
                target_column_name=self.target_column_name,
                training_data=self.training_data,
                validation_data=self.validation_data,
                validation_data_size=self.validation_data_size,
                weight_column_name=self.weight_column_name,
                cv_split_column_names=self.cv_split_column_names,
                n_cross_validations=self.n_cross_validations,
                test_data=self.test_data,
                test_data_size=self.test_data_size,
                featurization_settings=self._featurization._to_rest_object() if self._featurization else None,
                limit_settings=self._limits._to_rest_object() if self._limits else None,
                training_settings=self._training._to_rest_object() if self._training else None,
                primary_metric=self.primary_metric,
                log_verbosity=self.log_verbosity,
                forecasting_settings=None,
            )

        self._resolve_data_inputs(forecasting_task)
        self._validation_data_to_rest(forecasting_task)

        properties = RestAutoMLJob(
            display_name=self.display_name,
            description=self.description,
            experiment_name=self.experiment_name,
            tags=self.tags,
            compute_id=self.compute,
            properties=self.properties,
            environment_id=self.environment_id,
            environment_variables=self.environment_variables,
            services=self.services,
            outputs=to_rest_data_outputs(self.outputs),
            resources=self.resources,
            task_details=forecasting_task,
            identity=self.identity._to_job_rest_object() if self.identity else None,
            queue_settings=self.queue_settings,
        )

        result = JobBase(properties=properties)
        result.name = self.name
        return result

    @classmethod
    def _from_rest_object(cls, obj: JobBase) -> "ForecastingJob":
        properties: RestAutoMLJob = obj.properties
        task_details: RestForecasting = properties.task_details

        job_args_dict = {
            "id": obj.id,
            "name": obj.name,
            "description": properties.description,
            "tags": properties.tags,
            "properties": properties.properties,
            "experiment_name": properties.experiment_name,
            "services": properties.services,
            "status": properties.status,
            "creation_context": obj.system_data,
            "display_name": properties.display_name,
            "compute": properties.compute_id,
            "outputs": from_rest_data_outputs(properties.outputs),
            "resources": properties.resources,
            "identity": (
                _BaseJobIdentityConfiguration._from_rest_object(properties.identity) if properties.identity else None
            ),
            "queue_settings": properties.queue_settings,
        }

        forecasting_job = cls(
            target_column_name=task_details.target_column_name,
            training_data=task_details.training_data,
            validation_data=task_details.validation_data,
            validation_data_size=task_details.validation_data_size,
            weight_column_name=task_details.weight_column_name,
            cv_split_column_names=task_details.cv_split_column_names,
            n_cross_validations=task_details.n_cross_validations,
            test_data=task_details.test_data,
            test_data_size=task_details.test_data_size,
            featurization=(
                TabularFeaturizationSettings._from_rest_object(task_details.featurization_settings)
                if task_details.featurization_settings
                else None
            ),
            limits=(
                TabularLimitSettings._from_rest_object(task_details.limit_settings)
                if task_details.limit_settings
                else None
            ),
            training=(
                ForecastingTrainingSettings._from_rest_object(task_details.training_settings)
                if task_details.training_settings
                else None
            ),
            primary_metric=task_details.primary_metric,
            forecasting_settings=(
                ForecastingSettings._from_rest_object(task_details.forecasting_settings)
                if task_details.forecasting_settings
                else None
            ),
            log_verbosity=task_details.log_verbosity,
            **job_args_dict,
        )

        forecasting_job._restore_data_inputs()
        forecasting_job._validation_data_from_rest()

        return forecasting_job

    @classmethod
    def _load_from_dict(
        cls,
        data: Dict,
        context: Dict,
        additional_message: str,
        **kwargs: Any,
    ) -> "ForecastingJob":
        from azure.ai.ml._schema.automl.table_vertical.forecasting import AutoMLForecastingSchema
        from azure.ai.ml._schema.pipeline.automl_node import AutoMLForecastingNodeSchema

        if kwargs.pop("inside_pipeline", False):
            loaded_data = load_from_dict(AutoMLForecastingNodeSchema, data, context, additional_message, **kwargs)
        else:
            loaded_data = load_from_dict(AutoMLForecastingSchema, data, context, additional_message, **kwargs)
        job_instance = cls._create_instance_from_schema_dict(loaded_data)
        return job_instance

    @classmethod
    def _create_instance_from_schema_dict(cls, loaded_data: Dict) -> "ForecastingJob":
        loaded_data.pop(AutoMLConstants.TASK_TYPE_YAML, None)
        data_settings = {
            "training_data": loaded_data.pop("training_data"),
            "target_column_name": loaded_data.pop("target_column_name"),
            "weight_column_name": loaded_data.pop("weight_column_name", None),
            "validation_data": loaded_data.pop("validation_data", None),
            "validation_data_size": loaded_data.pop("validation_data_size", None),
            "cv_split_column_names": loaded_data.pop("cv_split_column_names", None),
            "n_cross_validations": loaded_data.pop("n_cross_validations", None),
            "test_data": loaded_data.pop("test_data", None),
            "test_data_size": loaded_data.pop("test_data_size", None),
        }
        job = ForecastingJob(**loaded_data)
        job.set_data(**data_settings)
        return job

    def _to_dict(self, inside_pipeline: bool = False) -> Dict:
        from azure.ai.ml._schema.automl.table_vertical.forecasting import AutoMLForecastingSchema
        from azure.ai.ml._schema.pipeline.automl_node import AutoMLForecastingNodeSchema

        schema_dict: dict = {}
        if inside_pipeline:
            schema_dict = AutoMLForecastingNodeSchema(context={BASE_PATH_CONTEXT_KEY: "./"}).dump(self)
        else:
            schema_dict = AutoMLForecastingSchema(context={BASE_PATH_CONTEXT_KEY: "./"}).dump(self)
        return schema_dict

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ForecastingJob):
            return NotImplemented

        if not super(ForecastingJob, self).__eq__(other):
            return False

        return self.primary_metric == other.primary_metric and self._forecasting_settings == other._forecasting_settings

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
# pylint: disable=protected-access, redefined-builtin
# disable redefined-builtin to use input as argument name
import re
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union

from pydash import get

from azure.ai.ml._utils.utils import is_data_binding_expression
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.constants._component import ComponentJobConstants
from azure.ai.ml.entities._inputs_outputs import Input, Output
from azure.ai.ml.entities._job.pipeline._io import PipelineInput, PipelineOutput
from azure.ai.ml.entities._job.sweep.search_space import Choice, Randint, SweepDistribution
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, JobException

# avoid circular import error
if TYPE_CHECKING:
    from azure.ai.ml.entities._builders import BaseNode
    from azure.ai.ml.entities._component.component import Component


class ComponentTranslatableMixin:
    _PYTHON_SDK_TYPE_MAPPING = {
        float: "number",
        int: "integer",
        bool: "boolean",
        str: "string",
    }

    @classmethod
    def _find_source_from_parent_inputs(cls, input: str, pipeline_job_inputs: dict) -> Tuple[str, Optional[str]]:
        """Find source type and mode of input/output from parent input.

        :param input: The input name
        :type input: str
        :param pipeline_job_inputs: The pipeline job inputs
        :type pipeline_job_inputs: dict
        :return: A 2-tuple of the type and the mode
        :rtype: Tuple[str, Optional[str]]
        """
        _input_name = input.split(".")[2][:-2]
        if _input_name not in pipeline_job_inputs.keys():
            msg = "Failed to find top level definition for input binding {}."
            raise JobException(
                message=msg.format(input),
                no_personal_data_message=msg.format("[input]"),
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )
        input_data = pipeline_job_inputs[_input_name]
        input_type = type(input_data)
        if input_type in cls._PYTHON_SDK_TYPE_MAPPING:
            return cls._PYTHON_SDK_TYPE_MAPPING[input_type], None
        return getattr(input_data, "type", AssetTypes.URI_FOLDER), getattr(input_data, "mode", None)

    @classmethod
    def _find_source_from_parent_outputs(cls, input: str, pipeline_job_outputs: dict) -> Tuple[str, Optional[str]]:
        """Find source type and mode of input/output from parent output.

        :param input: The input name
        :type input: str
        :param pipeline_job_outputs: The pipeline job outputs
        :type pipeline_job_outputs: dict
        :return: A 2-tuple of the type and the mode
        :rtype: Tuple[str, Optional[str]]
        """
        _output_name = input.split(".")[2][:-2]
        if _output_name not in pipeline_job_outputs.keys():
            msg = "Failed to find top level definition for output binding {}."
            raise JobException(
                message=msg.format(input),
                no_personal_data_message=msg.format("[input]"),
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )
        output_data = pipeline_job_outputs[_output_name]
        output_type = type(output_data)
        if output_type in cls._PYTHON_SDK_TYPE_MAPPING:
            return cls._PYTHON_SDK_TYPE_MAPPING[output_type], None
        if isinstance(output_data, dict):
            if "type" in output_data:
                output_data_type = output_data["type"]
            else:
                output_data_type = AssetTypes.URI_FOLDER
            if "mode" in output_data:
                output_data_mode = output_data["mode"]
            else:
                output_data_mode = None
            return output_data_type, output_data_mode
        return getattr(output_data, "type", AssetTypes.URI_FOLDER), getattr(output_data, "mode", None)

    @classmethod
    def _find_source_from_other_jobs(
        cls, input: str, jobs_dict: dict, pipeline_job_dict: dict
    ) -> Tuple[str, Optional[str]]:
        """Find source type and mode of input/output from other job.

        :param input: The input name
        :type input: str
        :param jobs_dict: The job dict
        :type jobs_dict:
        :param pipeline_job_dict: The pipeline job dict
        :type pipeline_job_dict: dict
        :return: A 2-tuple of the type and the mode
        :rtype: Tuple[str, Optional[str]]
        """
        from azure.ai.ml.entities import CommandJob
        from azure.ai.ml.entities._builders import BaseNode
        from azure.ai.ml.entities._job.automl.automl_job import AutoMLJob
        from azure.ai.ml.parallel import ParallelJob

        _input_regex = r"\${{parent.jobs.([^.]+).([^.]+).([^.]+)}}"
        m = re.match(_input_regex, input)
        if m is None:
            msg = "Failed to find top level definition for job binding {}."
            raise JobException(
                message=msg.format(input),
                no_personal_data_message=msg.format("[input]"),
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )
        _input_job_name, _io_type, _name = m.groups()
        _input_job = jobs_dict[_input_job_name]

        # we only support input of one job is from output of another output, but input mode should be decoupled with
        # output mode, so we always return None source_mode
        source_mode = None
        if isinstance(_input_job, BaseNode):
            # If source is base node, get type from io builder
            _source = _input_job[_io_type][_name]
            try:
                source_type = _source.type
                # Todo: get component type for registered component, and no need following codes
                # source_type is None means _input_job's component is registered component which results in its
                # input/output type is None.
                if source_type is None:
                    if _source._data is None:
                        # return default type if _input_job's output data is None
                        source_type = AssetTypes.URI_FOLDER
                    elif isinstance(_source._data, Output):
                        # if _input_job data is a Output object and we return its type.
                        source_type = _source._data.type
                    else:
                        #  otherwise _input_job's input/output is bound to pipeline input/output, we continue
                        #  infer the type according to _source._data. Will return corresponding pipeline
                        #  input/output type because we didn't get the component.
                        source_type, _ = cls._find_source_input_output_type(_source._data, pipeline_job_dict)
                return source_type, source_mode
            except AttributeError as e:
                msg = "Failed to get referenced component type {}."
                raise JobException(
                    message=msg.format(_input_regex),
                    no_personal_data_message=msg.format("[_input_regex]"),
                    target=ErrorTarget.PIPELINE,
                    error_category=ErrorCategory.USER_ERROR,
                ) from e
        if isinstance(_input_job, (CommandJob, ParallelJob)):
            # If source has not parsed to Command yet, infer type
            _source = get(_input_job, f"{_io_type}.{_name}")
            if isinstance(_source, str):
                source_type, _ = cls._find_source_input_output_type(_source, pipeline_job_dict)
                return source_type, source_mode
            return getattr(_source, "type", AssetTypes.URI_FOLDER), source_mode
        if isinstance(_input_job, AutoMLJob):
            # If source is AutoMLJob, only outputs is supported
            if _io_type != "outputs":
                msg = f"Only binding to AutoMLJob output is supported, currently got {_io_type}"
                raise JobException(
                    message=msg,
                    no_personal_data_message=msg,
                    target=ErrorTarget.PIPELINE,
                    error_category=ErrorCategory.USER_ERROR,
                )
            # AutoMLJob's output type can only be MLTABLE
            return AssetTypes.MLTABLE, source_mode
        msg = f"Unknown referenced source job type: {type(_input_job)}."
        raise JobException(
            message=msg,
            no_personal_data_message=msg,
            target=ErrorTarget.PIPELINE,
            error_category=ErrorCategory.USER_ERROR,
        )

    @classmethod
    def _find_source_input_output_type(cls, input: str, pipeline_job_dict: dict) -> Tuple[str, Optional[str]]:
        """Find source type and mode of input/output.

        :param input: The input binding
        :type input: str
        :param pipeline_job_dict: The pipeline job dict
        :type pipeline_job_dict: dict
        :return: A 2-tuple of the type and the mode
        :rtype: Tuple[str, Optional[str]]
        """
        pipeline_job_inputs = pipeline_job_dict.get("inputs", {})
        pipeline_job_outputs = pipeline_job_dict.get("outputs", {})
        jobs_dict = pipeline_job_dict.get("jobs", {})
        if is_data_binding_expression(input, ["parent", "inputs"]):
            return cls._find_source_from_parent_inputs(input, pipeline_job_inputs)
        if is_data_binding_expression(input, ["parent", "outputs"]):
            return cls._find_source_from_parent_outputs(input, pipeline_job_outputs)
        if is_data_binding_expression(input, ["parent", "jobs"]):
            try:
                return cls._find_source_from_other_jobs(input, jobs_dict, pipeline_job_dict)
            except JobException as e:
                raise e
            except Exception as e:
                msg = "Failed to find referenced source for input binding {}"
                raise JobException(
                    message=msg.format(input),
                    no_personal_data_message=msg.format("[input]"),
                    target=ErrorTarget.PIPELINE,
                    error_category=ErrorCategory.SYSTEM_ERROR,
                ) from e
        else:
            msg = "Job input in a pipeline can bind only to a job output or a pipeline input"
            raise JobException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )

    @classmethod
    def _to_input(
        cls,  # pylint: disable=unused-argument
        input: Union[Input, str, bool, int, float],
        pipeline_job_dict: Optional[dict] = None,
        **kwargs: Any,
    ) -> Input:
        """Convert a single job input value to component input.

        :param input: The input
        :type input: Union[Input, str, bool, int, float]
        :param pipeline_job_dict: The pipeline job dict
        :type pipeline_job_dict: Optional[dict]
        :return: The Component Input
        :rtype: Input
        """
        pipeline_job_dict = pipeline_job_dict or {}
        input_variable: Dict = {}

        if isinstance(input, str) and bool(re.search(ComponentJobConstants.INPUT_PATTERN, input)):
            # handle input bindings
            input_variable["type"], input_variable["mode"] = cls._find_source_input_output_type(
                input, pipeline_job_dict
            )

        elif isinstance(input, Input):
            input_variable = input._to_dict()
        elif isinstance(input, SweepDistribution):
            if isinstance(input, Choice):
                if input.values is not None:
                    input_variable["type"] = cls._PYTHON_SDK_TYPE_MAPPING[type(input.values[0])]
            elif isinstance(input, Randint):
                input_variable["type"] = cls._PYTHON_SDK_TYPE_MAPPING[int]
            else:
                input_variable["type"] = cls._PYTHON_SDK_TYPE_MAPPING[float]

            input_variable["optional"] = False
        elif type(input) in cls._PYTHON_SDK_TYPE_MAPPING:
            input_variable["type"] = cls._PYTHON_SDK_TYPE_MAPPING[type(input)]
            input_variable["default"] = input
        elif isinstance(input, PipelineInput):
            # Infer input type from input data
            input_variable = input._to_input()._to_dict()
        else:
            msg = "'{}' is not supported as component input, supported types are '{}'.".format(
                type(input), cls._PYTHON_SDK_TYPE_MAPPING.keys()
            )
            raise JobException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )
        return Input(**input_variable)

    @classmethod
    def _to_input_builder_function(cls, input: Union[Dict, SweepDistribution, Input, str, bool, int, float]) -> Input:
        input_variable = {}

        if isinstance(input, Input):
            input_variable = input._to_dict()
        elif isinstance(input, SweepDistribution):
            if isinstance(input, Choice):
                if input.values is not None:
                    input_variable["type"] = cls._PYTHON_SDK_TYPE_MAPPING[type(input.values[0])]
            elif isinstance(input, Randint):
                input_variable["type"] = cls._PYTHON_SDK_TYPE_MAPPING[int]
            else:
                input_variable["type"] = cls._PYTHON_SDK_TYPE_MAPPING[float]

            input_variable["optional"] = False
        else:
            input_variable["type"] = cls._PYTHON_SDK_TYPE_MAPPING[type(input)]
            input_variable["default"] = input
        return Input(**input_variable)

    @classmethod
    def _to_output(
        cls,  # pylint: disable=unused-argument
        output: Optional[Union[Output, Dict, str, bool, int, float]],
        pipeline_job_dict: Optional[dict] = None,
        **kwargs: Any,
    ) -> Output:
        """Translate output value to Output and infer component output type
        from linked pipeline output, its original type or default type.

        :param output: The output
        :type output: Union[Output, str, bool, int, float]
        :param pipeline_job_dict: The pipeline job dict
        :type pipeline_job_dict: Optional[dict]
        :return: The output object
        :rtype: Output
        """
        pipeline_job_dict = pipeline_job_dict or {}
        output_type = None
        if not pipeline_job_dict or output is None:
            try:
                output_type = output.type  # type: ignore
            except AttributeError:
                # default to url_folder if failed to get type
                output_type = AssetTypes.URI_FOLDER
            output_variable = {"type": output_type}
            return Output(**output_variable)
        output_variable = {}

        if isinstance(output, str) and bool(re.search(ComponentJobConstants.OUTPUT_PATTERN, output)):
            # handle output bindings
            output_variable["type"], output_variable["mode"] = cls._find_source_input_output_type(
                output, pipeline_job_dict
            )

        elif isinstance(output, Output):
            output_variable = output._to_dict()

        elif isinstance(output, PipelineOutput):
            output_variable = output._to_output()._to_dict()

        elif type(output) in cls._PYTHON_SDK_TYPE_MAPPING:
            output_variable["type"] = cls._PYTHON_SDK_TYPE_MAPPING[type(output)]
            output_variable["default"] = output
        else:
            msg = "'{}' is not supported as component output, supported types are '{}'.".format(
                type(output), cls._PYTHON_SDK_TYPE_MAPPING.keys()
            )
            raise JobException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )
        return Output(**output_variable)

    def _to_inputs(self, inputs: Optional[Dict], **kwargs: Any) -> Dict:
        """Translate inputs to Inputs.

        :param inputs: mapping from input name to input object.
        :type inputs: Dict[str, Union[Input, str, bool, int, float]]
        :return: mapping from input name to translated component input.
        :rtype: Dict[str, Input]
        """
        pipeline_job_dict = kwargs.get("pipeline_job_dict", {})
        translated_component_inputs = {}
        if inputs is not None:
            for io_name, io_value in inputs.items():
                translated_component_inputs[io_name] = self._to_input(io_value, pipeline_job_dict)
        return translated_component_inputs

    def _to_outputs(self, outputs: Optional[Dict], **kwargs: Any) -> Dict:
        """Translate outputs to Outputs.

        :param outputs: mapping from output name to output object.
        :type outputs: Dict[str, Output]
        :return: mapping from output name to translated component output.
        :rtype: Dict[str, Output]
        """
        # Translate outputs to Outputs.
        pipeline_job_dict = kwargs.get("pipeline_job_dict", {})
        translated_component_outputs = {}
        if outputs is not None:
            for output_name, output_value in outputs.items():
                translated_component_outputs[output_name] = self._to_output(output_value, pipeline_job_dict)
        return translated_component_outputs

    def _to_component(self, context: Optional[Dict] = None, **kwargs: Any) -> Union["Component", str]:
        """Translate to Component.

        :param context: The context
        :type context: Optional[context]
        :return: Translated Component.
        :rtype: Component
        """
        # Note: Source of translated component should be same with Job
        # And should be set after called _to_component/_to_node as job has no _source now.
        raise NotImplementedError()

    def _to_node(self, context: Optional[Dict] = None, **kwargs: Any) -> "BaseNode":
        """Translate to pipeline node.

        :param context: The context
        :type context: Optional[context]
        :return: Translated node.
        :rtype: BaseNode
        """
        # Note: Source of translated component should be same with Job
        # And should be set after called _to_component/_to_node as job has no _source now.
        raise NotImplementedError()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from inspect import Parameter
from typing import Dict, List, Optional, Union

from azure.ai.ml.constants._component import ExternalDataType
from azure.ai.ml.entities._inputs_outputs.utils import _remove_empty_values
from azure.ai.ml.entities._mixins import DictMixin, RestTranslatableMixin


class StoredProcedureParameter(DictMixin, RestTranslatableMixin):
    """Define a stored procedure parameter class for DataTransfer import database task.

    :keyword name: The name of the database stored procedure.
    :paramtype name: str
    :keyword value: The value of the database stored procedure.
    :paramtype value: str
    :keyword type: The type of the database stored procedure.
    :paramtype type: str
    """

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        value: Optional[str] = None,
        type: Optional[str] = None,  # pylint: disable=redefined-builtin
    ) -> None:
        self.type = type
        self.name = name
        self.value = value


class Database(DictMixin, RestTranslatableMixin):
    """Define a database class for a DataTransfer Component or Job.

    :keyword query: The SQL query to retrieve data from the database.
    :paramtype query: str
    :keyword table_name: The name of the database table.
    :paramtype table_name: str
    :keyword stored_procedure: The name of the stored procedure.
    :paramtype stored_procedure: str
    :keyword stored_procedure_params: The parameters for the stored procedure.
    :paramtype stored_procedure_params: List
    :keyword connection: The connection string for the database.
        The credential information should be stored in the connection.
    :paramtype connection: str
    :raises ~azure.ai.ml.exceptions.ValidationException: Raised if the Database object cannot be successfully validated.
        Details will be provided in the error message.

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_input_output_configurations.py
            :start-after: [START configure_database]
            :end-before: [END configure_database]
            :language: python
            :dedent: 8
            :caption: Create a database and querying a database table.
    """

    _EMPTY = Parameter.empty

    def __init__(
        self,
        *,
        query: Optional[str] = None,
        table_name: Optional[str] = None,
        stored_procedure: Optional[str] = None,
        stored_procedure_params: Optional[List[Dict]] = None,
        connection: Optional[str] = None,
    ) -> None:
        # As an annotation, it is not allowed to initialize the name.
        # The name will be updated by the annotated variable name.
        self.name = None
        self.type = ExternalDataType.DATABASE
        self.connection = connection
        self.query = query
        self.table_name = table_name
        self.stored_procedure = stored_procedure
        self.stored_procedure_params = stored_procedure_params

    def _to_dict(self, remove_name: bool = True) -> Dict:
        """Convert the Source object to a dict.

        :param remove_name: Whether to remove the `name` key from the  dict representation. Defaults to True.
        :type remove_name: bool
        :return: The dictionary representation of the class
        :rtype: Dict
        """
        keys = [
            "name",
            "type",
            "query",
            "stored_procedure",
            "stored_procedure_params",
            "connection",
            "table_name",
        ]
        if remove_name:
            keys.remove("name")
        result = {key: getattr(self, key) for key in keys}
        res: dict = _remove_empty_values(result)
        return res

    def _to_rest_object(self) -> Dict:
        # this is for component rest object when using Source as component inputs, as for job input usage,
        # rest object is generated by extracting Source's properties, see details in to_rest_dataset_literal_inputs()
        result = self._to_dict()
        return result

    def _update_name(self, name: str) -> None:
        self.name = name

    @classmethod
    def _from_rest_object(cls, obj: Dict) -> "Database":
        return Database(**obj)

    @property
    def stored_procedure_params(self) -> Optional[List]:
        """Get or set the parameters for the stored procedure.

        :return: The parameters for the stored procedure.
        :rtype: List[StoredProcedureParameter]
        """

        return self._stored_procedure_params

    @stored_procedure_params.setter
    def stored_procedure_params(self, value: Union[Dict[str, str], List, None]) -> None:
        """Set the parameters for the stored procedure.

        :param value: The parameters for the stored procedure.
        :type value: Union[Dict[str, str], StoredProcedureParameter, None]
        """
        if value is None:
            self._stored_procedure_params = value
        else:
            if not isinstance(value, list):
                value = [value]
            for index, item in enumerate(value):
                if isinstance(item, dict):
                    value[index] = StoredProcedureParameter(**item)
            self._stored_procedure_params = value


class FileSystem(DictMixin, RestTranslatableMixin):
    """Define a file system class of a DataTransfer Component or Job.

    e.g. source_s3 = FileSystem(path='s3://my_bucket/my_folder', connection='azureml:my_s3_connection')

    :param path: The path to which the input is pointing. Could be pointing to the path of file system. Default is None.
    :type path: str
    :param connection: Connection is workspace, we didn't support storage connection here, need leverage workspace
        connection to store these credential info. Default is None.
    :type connection: str
    :raises ~azure.ai.ml.exceptions.ValidationException: Raised if Source cannot be successfully validated.
        Details will be provided in the error message.
    """

    _EMPTY = Parameter.empty

    def __init__(
        self,
        *,
        path: Optional[str] = None,
        connection: Optional[str] = None,
    ) -> None:
        self.type = ExternalDataType.FILE_SYSTEM
        self.name: Optional[str] = None
        self.connection = connection
        self.path: Optional[str] = None

        if path is not None and not isinstance(path, str):
            # this logic will make dsl data binding expression working in the same way as yaml
            # it's written to handle InputOutputBase, but there will be loop import if we import InputOutputBase here
            self.path = str(path)
        else:
            self.path = path

    def _to_dict(self, remove_name: bool = True) -> Dict:
        """Convert the Source object to a dict.

        :param remove_name: Whether to remove the `name` key from the  dict representation. Defaults to True.
        :type remove_name: bool
        :return: The dictionary representation of the object
        :rtype: Dict
        """
        keys = ["name", "path", "type", "connection"]
        if remove_name:
            keys.remove("name")
        result = {key: getattr(self, key) for key in keys}
        res: dict = _remove_empty_values(result)
        return res

    def _to_rest_object(self) -> Dict:
        # this is for component rest object when using Source as component inputs, as for job input usage,
        # rest object is generated by extracting Source's properties, see details in to_rest_dataset_literal_inputs()
        result = self._to_dict()
        return result

    def _update_name(self, name: str) -> None:
        self.name = name

    @classmethod
    def _from_rest_object(cls, obj: Dict) -> "FileSystem":
        return FileSystem(**obj)
